<?php
/**
 * Self-documenting extension to the PHP-XMLRPC server
 * Mantis-augmented version
 *
 * @version $Id$
 * @copyright (C) G. Giunta 2008
 * @license code licensed under the GPL v2 License
 */

	// requires: xmlrpc.inc, xmlrpcs.inc (version 2.0RC3 or later), docxmlrpcs.inc

	class mantis_xmlrpc_server extends documenting_xmlrpc_server
	{
    	// keep the same API as the register function of the soap equivalent
    	function register( $funcname, $funcparams, $returntype, $namespace, $x=null, $y=null, $z=null, $docs='' )
        {
            $sig = array( mantis_xmlrpc_server::xsd2xmlrpcType( $returntype['return'] ) );
            $pdocs = array( $returntype['return'] );
            foreach( $funcparams as $param )
            {
                $sig[] = mantis_xmlrpc_server::xsd2xmlrpcType( $param );
                $pdocs[] = $param;
            }
    	    $this->add_to_map( $funcname, $funcname, array( $sig ), $docs, array( $pdocs ) );
    	}

        // api compat: this code is called by the mantisconnect error handler
        /// @todo: if faultdetail === '', we should use debug, otherwise use that
        function fault($faultcode, $faultstring, $faultactor='', $faultdetail='')
        {
            if ( $faultcode == 0 ) // we never know, we are possibly getting a string
            {
                $faultcode = $GLOBALS['xmlrpcerruser'];
            }
            $this->fault = new xmlrpcresp( null, $faultcode, $faultstring );
        }

        // api compat:
        // hook in the middle of the server execution path, to store more info
        // (inside the server itself) about the current respose being built.
        // Later the exception handler might call send_response, that will finish
        // the job using the info we just stocked
        function parseRequestHeaders(&$data, &$req_encoding, &$resp_encoding, &$resp_compression)
		{
    		$r = parent::parseRequestHeaders($data, $req_encoding, $resp_encoding, $resp_compression);
    		$this->resp_encoding = $resp_encoding;
    		$this->resp_compression = $resp_compression;
    		return $r;
		}

        // api compat: this code is called by the mantisconnect error handler
        function send_response()
        {
            // var name swapping is correct!
            $resp_charset = isset( $this->resp_encoding ) ? $this->resp_encoding : (
                $this->response_charset_encoding == 'auto' ? '' : $this->response_charset_encoding );
            $resp_encoding = isset( $this->resp_compression ) ? $this->resp_compression : '';
            $r = isset( $this->fault ) ? $this->fault : new xmlrpcresp( null, $GLOBALS['xmlrpcerruser'], 'Internal server error: send_response() invoked without a previous call to fault()' );
            $return_payload = false;

            // code copied and pasted from parent::service()
            // it would be mich nicer if the base class refactored this out...

			if($this->debug > 2 && $GLOBALS['_xmlrpcs_occurred_errors'])
			{
				$this->debugmsg("+++PROCESSING ERRORS AND WARNINGS+++\n" .
					$GLOBALS['_xmlrpcs_occurred_errors'] . "+++END+++");
			}

			$payload=$this->xml_header($resp_charset);
			if($this->debug > 0)
			{
				$payload = $payload . $this->serializeDebug($resp_charset);
			}

			// G. Giunta 2006-01-27: do not create response serialization if it has
			// already happened. Helps building json magic
			if (empty($r->payload))
			{
				$r->serialize($resp_charset);
			}
			$payload = $payload . $r->payload;

			if ($return_payload)
			{
				return $payload;
			}

			// if we get a warning/error that has output some text before here, then we cannot
			// add a new header. We cannot say we are sending xml, either...
			if(!headers_sent())
			{
				header('Content-Type: '.$r->content_type);
				// we do not know if client actually told us an accepted charset, but if he did
				// we have to tell him what we did
				header("Vary: Accept-Charset");

				// http compression of output: only
				// if we can do it, and we want to do it, and client asked us to,
				// and php ini settings do not force it already
				$php_no_self_compress = !ini_get('zlib.output_compression') && (ini_get('output_handler') != 'ob_gzhandler');
				if($this->compress_response && function_exists('gzencode') && $resp_encoding != ''
					&& $php_no_self_compress)
				{
					if(strpos($resp_encoding, 'gzip') !== false)
					{
						$payload = gzencode($payload);
						header("Content-Encoding: gzip");
						header("Vary: Accept-Encoding");
					}
					elseif (strpos($resp_encoding, 'deflate') !== false)
					{
						$payload = gzcompress($payload);
						header("Content-Encoding: deflate");
						header("Vary: Accept-Encoding");
					}
				}

				// do not ouput content-length header if php is compressing output for us:
				// it will mess up measurements
				if($php_no_self_compress)
				{
					header('Content-Length: ' . (int)strlen($payload));
				}
			}
			else
			{
				error_log('XML-RPC: xmlrpc_server::service: http headers already sent before response is fully generated. Check for php warning or error messages');
			}

			print $payload;
        }

        // api compat: mantisconnect sometimes returns SoapFault objects, that
		// get encoded as xmlrpc structs. We recode them to xmlrpc errors.
		function execute($m, $params=null, $paramtypes=null)
		{
    		// reset internal error placeholder before every request parsing
    		$this->fault = null;
    		$r = parent::execute( $m, $params, $paramtypes );
    		if ( $r->faultCode() == 0 )
    		{
    		    $v = $r->value();
    		    if ( $v->kindOf() == 'struct' && $v->structmemexists( 'faultcode' ) && $v->structmemexists( 'faultstring' ) && $v->structmemexists( 'faultactor' ) && $v->structmemexists( 'faultdetail' ) )
    		    {
    		        // this is a soap error. we intercept it and recode it in an xmlrpc one
    		        $faultcode = $v->structmem( 'faultcode' );
                    $faultcode = $faultcode->scalarval();
    		        $faultstring = $v->structmem( 'faultstring' );
    		        $faultstring = $faultstring->scalarval();
    		        if ( $faultcode == 0 ) // we never know, we are possibly getting a string
                    {
                        $faultcode = $GLOBALS['xmlrpcerruser'];
                    }
                    $r = new xmlrpcresp( null, $faultcode, $faultstring );
    		    }
    		}
    		return $r;
		}

        // utility function
        static function xsd2xmlrpcType( $type )
        {
            if ( preg_match( '#^tns:#', $type ) )
            {
                return 'struct';
            }
            $type = preg_replace( '#^xsd:#', '', $type );
            switch( $type )
            {
                case 'base64Binary':
                    return 'base64';
                case 'integer':
                    return 'int';
                /*// voluntary fallthrough
                case 'ProjectData':
                case 'ProjectArray':
                case 'ObjectRefArray':
                case 'IssueData':
                case 'AccountDataArray':
                case 'IssueDataArray':
                case 'IssueHeaderDataArray':
                case 'IssueNoteData':
                case 'RelationshipData':
                    return 'struct';*/
                default:
                    // here we could actually filter limiting to the types we know...
                    return $type;
            }
        }
	}

?>